#!/bin/sh
#
# kernelmod.sh
# Copyright (C) 2003  Richard de Vroede <richard@fenix.getmyip.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# Your copy is available at: http://www.gnu.org/licenses/gpl.html

# version, also change Makefile
version="0.8.4"

patchdir=$(echo $0 | rev | cut -d "/" -f2- | rev)
cd $patchdir
kmoddir=`pwd`
debug=0
BSD=1

# command dependencies
PTCH=$(which patch || exit 1)
CC=$(which cc || exit 1)
MODPRB=$(which modprobe || exit 1)
LD=$(which ld || exit 1)
TR=$(which tr || exit 1)
UNAME=$(which uname || exit 1)

setkernelver() {
  if [ -n "$KSOURCE" ]; then
    VERSION=`head -n 4 $KSOURCE/Makefile | grep VERSION | awk -F' = ' '{print $2}'`
    PATCHLEVEL=`head -n 4 $KSOURCE/Makefile | grep PATCHLEVEL | awk -F' = ' '{print $2}'`
    SUBLEVEL=`head -n 4 $KSOURCE/Makefile | grep SUBLEVEL | awk -F ' = ' '{print $2}'`
    EXTRAVERSION=`head -n 4 $KSOURCE/Makefile | grep EXTRAVERSION | awk -F ' = ' '{print $2}'`
    if [ -n "$EXTRAVERSION" ]; then
      k_ver="$VERSION.$PATCHLEVEL.$SUBLEVEL-$EXTRAVERSION"
    else
      k_ver="$VERSION.$PATCHLEVEL.$SUBLEVEL"
    fi
    echo "Using kernel source dir to get kernel version ($k_ver)"
  else
    k_ver=$(cat /proc/sys/kernel/osrelease)
    echo "Using running kernel to get kernel version ($k_ver)"
  fi
  maj_ver=$(echo $k_ver | awk -F "." '{print $1 "." $2}')
  min_ver=$(echo $k_ver | cut -d "." -f3 | cut -d "-" -f1)
  release=$(echo $k_ver | cut -s -d "-" -f2)
}


os=$(uname)
if [ -f /etc/redhat-release ]; then
   DISTRO="RedHat"
   DVER=$(cat /etc/redhat-release | cut -d " " -f5-)
elif [ -f /etc/SuSE-release ]; then 
   DISTRO="SuSE"
   DVER=$(cat /etc/SuSE-release | head -n1 | cut -d " " -f3)
elif [ -f /etc/debian_version ]; then 
   DISTRO="Debian"
   DVER=$(cat /etc/debian_version)
elif [ -f /etc/slackware-version ]; then
   DISTRO="Slackware"
   DVER=$(cat /etc/slackware-version | cut -d " " -f2)
else
   DISTRO=$os
fi

if [ "$debug" == "1" ]; then
    KMODLOG=$kmoddir/debug.log
else
    KMODLOG=/dev/null
fi

header() {
    echo "-------------------------------------------------"
    echo "  kernelmod v$version"
    echo "  Richard de Vroede <richard@oip.tudelft.nl>"
    echo "  James Cameron <quozl@us.netrek.org>"
    echo "-------------------------------------------------"
}

function menu() {
 DIALOG=`which dialog 2>/dev/null || echo ""`
 if [ -z "$DIALOG" ]; then
  echo "menuinstall needs dialog installed."
  exit 1
 fi
 num=0; cd src
 for module in `ls -d * | grep -v CVS | grep -v README`; do num=$(($num+1)); done
 cd -
 echo -e "$DIALOG --title \"Select Modules\" \\
  --backtitle \"kernelmod $version Module Selection\" \\
  --checklist \\
  \"Please select the modules you want to build\" 10 70 $num \\" >/tmp/kmodmenu
 no=0; cd src
 for module in `ls -d * | grep -v CVS | grep -v README`; do
  desc=$(grep ^desc: $module/meta.kmod | cut -d ":" -f2)
  if [ ! -z "$desc" ]; then
    no=$(($no+1))
    if [ $no -lt $num ];then
     echo " $module \"$desc\" 0 \\" >> /tmp/kmodmenu
    else
     echo " $module \"$desc\" 0" >> /tmp/kmodmenu
    fi
  fi
 done
 cd -
 . /tmp/kmodmenu 2>/tmp/kmodchoices
 clear
 MODULES=$(cat /tmp/kmodchoices | sed 's/\"//g')
 if [ -z "$MODULES" ]; then exit 0; fi
 rm -f /tmp/kmod*
}

findsource() {
   echo "--> Finding source for your running kernel."

   # Easiest way to locate the 2.6 sources, also works for some 2.4 kernels
   KSOURCE="/lib/modules/$k_ver/build"
   if [ -d $KSOURCE ]; then FOUND=1; else KSOURCE=""; fi
  
   if [ -z "$FOUND" ]; then
      if [ "$DISTRO" = "RedHat" -a -d "/usr/src/linux-$k_ver" ]; then
         KSOURCE="/usr/src/linux-$k_ver"
      fi
      if [ "$DISTRO" = "Debian" ]; then
         if [ -d "/usr/src/kernel-source-$maj_ver.$min_ver" ]; then
            KSOURCE="/usr/src/kernel-source-$maj_ver.$min_ver"
            echo "Found $KSOURCE, this probably is what we need."
            FOUND=1
         else
            echo "/usr/src/kernel-source-$maj_ver.$min_ver not found."
            echo "Install the kernel-source belonging to your running kernel!"
            echo "apt-get install kernel-source-$maj_ver.$min_ver"
      	    echo "cd /usr/src/"
   	    echo "tar -xjf kernel-source-$maj_ver.$min_ver.tar.bz2"
            exit 1
         fi
      fi
      if [ ! -z "$KSOURCE" ]; then
         # Found a candidate, check the version.
         TEMP1=$(cat $KSOURCE/include/linux/version.h 2>$KMODLOG | grep `$UNAME -r`)
         if [ ! -z "TEMP1" ]; then FOUND=1; fi
      fi
      if [ -z "$FOUND" ]; then
         # Not found yet.
         for lindir in `ls -1d /usr/src/linux*`; do
            # Is lindir a symlink?
            TEMP=$(file $lindir | cut -d " " -f2)
            if [ "$TEMP" = "symbolic" -o "$TEMP" = "broken" ]; then
               # Yes, skip it.
               echo "$lindir is a symlink. Skipping it."
            else
               # No, check the version.
               TEMP1=$(cat $lindir/include/linux/version.h 2>$KMODLOG | grep `$UNAME -r`)
               if [ -n "$TEMP1" -o "$TEMP1" != "" ]; then
                  # Found it. 
                  KSOURCE=$lindir
                  FOUND=1
               fi
            fi
         done
      fi
   fi

   if [ ! -z "$FOUND" ]; then
       echo "Found the source for your running kernel."
       echo "Location: $KSOURCE"
       return
   else
      echo "Didn't find it. Cowardly bailing out."
      exit 1
   fi
}

rollbackpatch() {
 if [ -z "$1" ]; then
  if [ "$BSD" ]; then
   module="bsd-mppe"
  else
   module="openssl-mppe"
  fi
 else
  module="$1"
 fi
 patchdir="$kmoddir/src/$module/$maj_ver"
 echo "--> Rolling back patches for $module."
 if [ -z "$KSOURCE" ]; then 
  findsource
 elif [ -z "$error" ]; then
  echo "You specified the location of the kernel source."
  echo "I trust your judgement, so I won't verify."
 fi
 
 # 2.6 specific logic to handle patch reversal
 if [ ! "$maj_ver" = "2.6" ]; then
    echo "-------------------------------------------------"
    for patch in `ls -1 $patchdir/*-$maj_ver.$min_ver-*.patch`; do
     echo "Reverting patch $patch."
     echo "-------------------------------------------------"
     for file in `cat $patch | grep "\-\-\- " | awk -F " " '{print $2}' | sed 's/.orig//g' | cut -d "/" -f2-`; do
      if [ -f "$KSOURCE/$file.bak" ]; then
       echo -n "Processing file $file..."
       mv -f "$KSOURCE/$file.bak" "$KSOURCE/$file"
       echo "Done."
      fi
     done
    done
  else
    echo "No patches to revert for Linux 2.6"
    cd $patchdir;
  fi
}

version() {
cat <<EOF
kernelmod v$version
EOF
}

usage() {
   header
cat <<EOF
   Usage: $0 [-a architecture] [-e emulation] [-h] [-k kernel-source] [-m] [-o] [-r module] [-s] [-v]

   Options:
      -a    Define architecture (Default: i386)
      -e    Define emulation (Supported emulations: elf_i386 i386linux
            elf_i386_glibc21; Default: elf_i386)
      -h    Shows this help
      -k    The location of the kernel source
      -m    Use menuinstall
      -o    Use OpenSSL licensed module instead of BSD licensed module
            (you need ppp < 2.4.2 to use this)
      -p    patch only, do not compile
      -r    Revert the patches
      -s    Build for SMP (Default: No SMP)
      -v    Show version information
EOF
}

while getopts ":a:e:hk:mopr:sv" opt; do
   case $opt in
      a) arch=$OPTARG ;;
      e) emulation=$OPTARG ;;
      h) usage; exit ;;
      k) KSOURCE=$OPTARG ;;
      m) MENU=1; menu ;;
      o) patchdir="$patchdir/openssl"; BSD=0 ;;
      p) patchonly="1" ;;
      r) MODULES=$OPTARG; header; rollbackpatch; exit ;;
      s) smp="1" ;;
      v) version; exit ;;
     \?) echo "Wrong option(s)"; usage; exit 1 ;;
   esac
done

shift $(($OPTIND-1))

header
setkernelver
echo "--> Summary."
echo "OS:             $os"

echo "Distribution:   $DISTRO $DVER"

echo "Kernel version: $k_ver"

if [ -z "$arch" ]; then 
    arch=$($UNAME -p | $TR A-Z a-z)
fi
echo "Architecture:   $arch"

# Supported emulations: elf_i386 i386linux elf_i386_glibc21
if [ -z "$emulation" ]; then 
    emulation="elf_i386"
fi
echo "Emulation:      $emulation"

echo -n "SMP:            "
if [ -n "$smp" ]; then echo "Yes"; else echo "No"; fi

echo "-------------------------------------------------"
echo "--> Preparing system."
mkdir -p $kmoddir/build
echo "Clearing out builddir $kmoddir/build"
rm -rf $kmoddir/build/*

if [ -z "$KSOURCE" ]; then 
   findsource
elif [ -n "$patchonly" ]; then
   # if we are only patching then don't mangle the kernel 
   echo "I trust your judgement, so I won't verify."
else
   echo "You specified the location of the kernel source."
   if [ ! -d "$KSOURCE" ]; then
      echo "It's not there!!!!"
      exit 1
   fi
   echo "I trust your judgement, so I won't verify."
   echo "I will set the correct kernel version though."
   VERSION=$(echo $maj_ver | cut -d . -f1)
   PATCHLEVEL=$(echo $maj_ver | cut -d . -f2)
   SUBLEVEL=$min_ver
   if [ -n "$release" ]; then
       EXTRAVERSION="-$release"
   else
       EXTRAVERSION=""
   fi
   if [ ! "$maj_ver" = "2.6" ]; then
   	LINES=$(wc -l $KSOURCE/Makefile | awk -F " " '{print $1}')
   	LINES=$(($LINES-4))
   	tail -n $LINES $KSOURCE/Makefile > /tmp/Makefile
   	echo -e "VERSION = $VERSION\nPATCHLEVEL = $PATCHLEVEL\nSUBLEVEL = $SUBLEVEL\nEXTRAVERSION = $EXTRAVERSION\n" > $KSOURCE/Makefile
   	cat /tmp/Makefile >> $KSOURCE/Makefile
   	make -s symlinks -C $KSOURCE
   	make -s include/linux/version.h -C $KSOURCE
   fi
fi

if [ "$maj_ver" = "2.6" ]; then
	cd $maj_ver
	# This process copies rather than patches the installed kernel sources.
	echo "-----------------------------------------------------"
	echo "You're using the 2.6 version of the LINUX kernel"
	echo "Copying in the $maj_ver.$min_ver kernel sources"
	echo "-----------------------------------------------------"
	make_args="-C $KSOURCE M=$kmoddir/$maj_ver"
	make $make_args clean 2>&1 > /dev/null
	# Need to copy in ppp-comp.h from kernel source
	if [ ! -f $KSOURCE/include/linux/ppp-comp.h ]; then
		echo "Cannot find ppp-comp.h - missing from kernel source?"
		exit 1
	else
		echo "Copying ppp-comp.h from $KSOURCE/include/linux"
		cp $KSOURCE/include/linux/ppp-comp.h .
	fi
	# Need to copy in ppp-comp.h from kernel source
	# This is in the kernel-source RPM on FC2, unlike ppp-comp.h, which is in the kernel RPM
	if [ ! -f $KSOURCE/drivers/net/ppp_generic.c ]; then
		if [ ! -f /usr/src/linux-$k_ver/drivers/net/ppp_generic.c ]; then
			echo "Cannot find ppp_generic.c - missing from kernel source?"
			rm -f ppp-comp.h
			exit 1
		else
			echo "Copying ppp_generic.c from /usr/src/linux-$k_ver/drivers/net"
			cp /usr/src/linux-$k_ver/drivers/net/ppp_generic.c .
		fi
	else
		echo "Copying ppp_generic.c from $KSOURCE/drivers/net"
		cp $KSOURCE/drivers/net/ppp_generic.c .
	fi
	for patch in linux-2.6.2-pad.patch 02-ppp_generic.c.patch \
			03-ppp_mppe_compress.c.patch 04-ppp_mppe_compress.c_license.patch
	do
		patch -p1 < $patch
	done
	echo "-----------------------------------------------------"
	echo "Building the $maj_ver.$min_ver specific mppe modules"
	echo "-----------------------------------------------------"
	error=""
	make $make_args modules || error=1
	if [ -z "$error" ]; then
		echo "-----------------------------------------------------"
		echo "Installing the $maj_ver.$min_ver specific mppe modules"
		echo "-----------------------------------------------------"
		strip -g ppp_generic.ko
		strip -g ppp_mppe.ko
		install -m 0744 ppp_generic.ko /lib/modules/$k_ver/kernel/drivers/net/ppp_generic_mppe.ko
		install -m 0744 ppp_mppe.ko /lib/modules/$k_ver/kernel/drivers/net/ppp_mppe.ko
		if [ ! -f /lib/modules/$k_ver/kernel/drivers/net/ppp_generic_original.ko ]; then
			mv -f /lib/modules/$k_ver/kernel/drivers/net/ppp_generic.ko \
				/lib/modules/$k_ver/kernel/drivers/net/ppp_generic_original.ko
		fi
		rm -f /lib/modules/$k_ver/kernel/drivers/net/ppp_generic.ko
		cp -p /lib/modules/$k_ver/kernel/drivers/net/ppp_generic_mppe.ko \
			/lib/modules/$k_ver/kernel/drivers/net/ppp_generic.ko
		[ -f /boot/System.map-$k_ver ] && depmod -a -F /boot/System.map-$k_ver
	else
		# Whoops, something went wrong
		echo "-----------------------------------------------------"
		echo "Error during module build - reverting patches"
		echo "-----------------------------------------------------"
		for patch in 04-ppp_mppe_compress.c_license.patch 03-ppp_mppe_compress.c.patch \
			02-ppp_generic.c.patch linux-2.6.2-pad.patch
		do
			patch -p1 -R < $patch
		done
		rm -f ppp_generic.c ppp-comp.h
		make $make_args clean
		exit 1
	fi
	cd $patchdir
else
	echo "--> Making sure the kernel source is configured."
	if [ ! -f "$KSOURCE/.config" ]; then 
	   cd $KSOURCE || (echo "$0: error, no source directory" ; exit 1)
	if [ "$DISTRO" = "RedHat" ]; then
	   # Does not exist in standard kernel. Have to make a patch for the Makefile.
	   make oldconfig_nonint 2>$KMODLOG 1>$KMODLOG || echo "Failed!"
	else
          if [ ! -n "$patchonly" ]; then
            make oldconfig || echo "Failed!"
          fi
	fi
	   cd $patchdir
	   echo "Done."
	else
	   echo "Found $KSOURCE/.config, so it probably is configured."
	fi
	echo "--> Looking for kernel-headers and a good modversions.h"
#       echo "Updating dependencies."
#       make dep 2>$KMODLOG 1>$KMODLOG || echo "Failed!"
	# This breaks building of modules on RedHat, Debian and maybe others. 
	# Needed to make modversions.h, so let's do something with
	# kernel-headers for Debian.
	if [ "$DISTRO" = "Debian" ]; then
  	   if [ -d "/usr/src/kernel-headers-$k_ver" ]; then
	      KHEADERS="/usr/src/kernel-headers-$k_ver"
	      echo "Found $KHEADERS, this probably is what we need."
	   else
	      echo "/usr/src/kernel-headers-$k_ver not found."
	      echo "Install the kernel-headers belonging to your running kernel!"
	      echo "apt-get install kernel-headers-$k_ver"
	      exit 1
	   fi
	fi
	if [ ! -z "$KHEADERS" ]; then
	   KINCLUDE=$(echo "$KHEADERS/include")
	else
	   KINCLUDE=$(echo "$KSOURCE/include")
	fi

	# RedHat 6.2, kernel 2.2.x
	if [ -f "$KSOURCE/include/config/modversions.h" -a ! -f $KSOURCE/include/linux/modversions.h ]; then
	   ln -s $KSOURCE/include/config/modversions.h $KSOURCE/include/linux/modversions.h
	fi

	echo "--> Done."
	echo "-------------------------------------------------"

	if [ ! "$MENU" ]; then
	 MODULES="bsd-mppe"
	fi

	 for module in $MODULES; do
	  license=$(grep ^license: $kmoddir/src/$module/meta.kmod | cut -d ":" -f2)
	  echo "================================================="
	  echo "--> Building module $module."
	  echo "License:        $license"
	  echo "================================================="
	  echo "--> Locating patches."
	  patchdir="$kmoddir/src/$module/$maj_ver"
	  if [ -d "$patchdir" ]; then
	   DEFAULTS=$(ls -1 $patchdir/*-default-*.patch)
	   if [ -n "$DEFAULTS" ]; then
	    echo "Found patches for $maj_ver."
	    echo "Checking for specific patches."
	    for patch in `ls -1 $patchdir/*-default-*.patch`; do
	     pre=$(echo $patch | rev | cut -d "/" -f1 | rev | awk -F "-default-" '{print $1}')
	     post=$(echo $patch | rev | cut -d "/" -f1 | rev  | awk -F "-default-" '{print $2}' | sed 's/\.patch//g')
	     if [ ! -f "$patchdir/$pre-$maj_ver.$min_ver-$post.patch" ]; then
	      cd $patchdir/
	      ln -s $pre-default-$post.patch $pre-$maj_ver.$min_ver-$post.patch
	      cd -
	     fi
	    done
	   fi
	   PATCHES=$(ls -1 $patchdir/*-$maj_ver.$min_ver-*.patch)
	   if [ -n "$PATCHES" ]; then
	    echo "Found patches for $maj_ver.$min_ver"
	    echo "-------------------------------------------------"
	    for patch in `ls -1 $patchdir/*-$maj_ver.$min_ver-*.patch`; do
	     echo "Applying patch $patch"
	     $PTCH -p1 -N -b -z .bak -d "$KSOURCE" -i $patch || error=1
	     if [ "$error" = "1" ]; then
	      echo "Found problems in module(s) from patch $patch."
	      rollbackpatch $module
	      exit 1
	     fi
	     for hfile in `cat $patch | grep "\-\-\- " | grep "\.h" | awk '{print $2}' | sed 's/.orig//g' | cut -d "/" -f2-`; do
	      cp -f "$KSOURCE/$hfile" $kmoddir/build
	      if [ ! -z "$KHEADERS" ]; then
	       LOC=$(echo $hfile | rev | cut -d "/" -f2- | rev)
	       if [ -d "$KHEADERS/$LOC" ]; then
	        echo "Copying $hfile to $KHEADERS"
	        cp -f "$KSOURCE/$hfile" $KHEADERS/$hfile
	       fi
	      fi
	     done
	     for cfile in `cat $patch | grep "\-\-\- " | grep "\.c" | awk '{print $2}' | sed 's/.orig//g' | cut -d "/" -f2-`; do
	      cp -f "$KSOURCE/$cfile" $kmoddir/build
	     done
	    done
	   fi
	  fi
	  echo "Copying extra sources to $KSOURCE"
	  for srcfile in `cat $kmoddir/src/$module/meta.kmod | grep -v "^#" | grep "^index\:"`; do
	   srcfilebase=$(echo $srcfile | cut -d ":" -f2- | rev | cut -d "/" -f1 | rev)
	   srcfilename=$(echo $srcfile | cut -d ":" -f2-)
	   echo "$srcfilebase --> $KSOURCE/$srcfilename"
	   cp -f "$kmoddir/src/$module/$srcfilebase" $KSOURCE/$srcfilename
	  done
	  echo "Copying extra sources to $kmoddir/build/"
	  cp -f "$kmoddir/src/$module/"*.[ch] $kmoddir/build/ 2>$KMODLOG 1>$KMODLOG
	  echo "-------------------------------------------------"
          if [ -n "$patchonly" ]; then
            exit 0
          fi
	  echo "--> Building kernel modules."
	  cd $kmoddir/build
	# [mod1]for cfile in `cat $patch | grep "\-\-\- " | grep "\.c" | awk -F " " '{print $2}' | sed 's/.orig//g' | cut -d "/" -f2-`; do
	  for cfile in `ls -1 *.c`; do
	   basename=$(echo $cfile | rev | cut -d "/" -f1 | sed 's/^c.//g' | rev)
	   excludes=$(cat $kmoddir/src/$module/meta.kmod | grep -v "^#" | grep "^excludes\:")
	   TEMP=$(echo $excludes | grep "$basename")
	   if [ -z "$TEMP" ]; then
	    echo "Building module $basename.o"
	    CFLAGS=" -D__KERNEL__"
	    CFLAGS=$(echo "$CFLAGS -D__KERNEL_MODULE_$arch=1")
	    if [ -n "$smp" ]; then CFLAGS=$(echo "$CFLAGS -D__BOOT_KERNEL_SMP=1"); fi
	    CFLAGS=$(echo "$CFLAGS -Wall -Wstrict-prototypes -Wno-trigraphs -O2 -fomit-frame-pointer")
	    CFLAGS=$(echo "$CFLAGS -fno-strict-aliasing -fno-common -Wno-unused -pipe")
	    if [ "$maj_ver" != "2.2" -a "$arch" != "ia64" ]; then
	     CFLAGS=$(echo "$CFLAGS -mpreferred-stack-boundary=2")
	    fi
	    CFLAGS=$(echo "$CFLAGS -I$KINCLUDE")
	    CFLAGS=$(echo "$CFLAGS -include $KINCLUDE/linux/modversions.h")
	    CFLAGS=$(echo "$CFLAGS -march=$arch -DMODULE -DMODVERSIONS")
	    CFLAGS=$(echo "$CFLAGS -DKBUILD_BASENAME=$basename -DEXPORT_SYMTAB")
	    $CC $CFLAGS -c $basename.c || error=1
	   fi
	  done
	  groups=$(cat $kmoddir/src/$module/meta.kmod | grep -v "^#" | grep "^ld\:" | cut -d ":" -f2)
	  for group in `echo $groups`; do
	   mods=$(cat $kmoddir/src/$module/meta.kmod | grep -v "^#" | grep "^ld\:$group\:" | cut -d ":" -f3)
	   echo "Building module $group"
	   $LD -m $emulation -A $arch -r -o $group $mods
	   rm -f $mods
	  done
	  echo "-------------------------------------------------"
	  if [ "$error" = "1" ]; then
	   echo "--> There has been an error. Leaving buildstuff in $kmoddir/build"
	   rollbackpatch $module
	   exit 1
	  else
	   for destfile in `cat $kmoddir/src/$module/meta.kmod | grep -v "^#" | grep "^dest\:"`; do
	    basename=$(echo $destfile | rev | cut -d "/" -f1 | sed 's/^o.//g' | rev)
	    TEMP=`eval "echo $mods"`
	    TEMP=$(echo $TEMP | grep $basename | cut -d "." -f1)
	    ddir=$(echo $destfile | cut -d ":" -f2- | rev | cut -d "/" -f2- | rev)
	    if [ "$maj_ver" = "2.4" ]; then ddir=$(echo "kernel/$ddir"); fi
	    echo "Installing module $basename.o in /lib/modules/$k_ver/$ddir/"
	    echo $basename >> /tmp/kmodlist
	    mv -f $kmoddir/build/$basename.o /lib/modules/$k_ver/$ddir/
	   done
	   echo "--> $module seems OK. Removing $kmoddir/build"
	   rm -rf $kmoddir/build
	  fi
	  error=0
	  echo "================================================="
	 done

	echo "Updating module dependencies"
	if [ -f "/tmp/kmodlist" ]; then
	 depmod -a
	 for module in `cat /tmp/kmodlist`; do
	  rmmod -r $module 2>$KMODLOG
	  modprobe $module 2>$KMODLOG 1>$KMODLOG
	  TEMP=$(lsmod | grep $module)
	  if [ -z "$TEMP" ]; then 
	   echo "WARNING: $module not loaded! Must be something wrong:"
	   modprobe $module
	   error=1
	  fi
	 done
	fi
fi

rm -f /tmp/kmodlist
if [ "$error" = "1" ]; then
    echo "!! There has been an error. Try with debugging enabled."
    exit 1
else
    echo "!! Everything seems OK."
fi

if [ "$error" = "1" ]; then exit 1; fi

cd $kmoddir
if [ "$DISTRO" = "RedHat" ]; then
    echo "-------------------------------------------------"
    echo "You are running RedHat, do you want me to make an RPM for you?"
    echo "Note: the modules are already installed, this RPM is for future use"
    echo "      or distribution."
    echo -n "Make an RPM? (yes/no) > "
    read reply
    if [ "$reply" = "yes" -o "$reply" = "y" -o "$reply" = "Y" ]; then
        cp -f confmod.sh /usr/src/redhat/SOURCES/
        rpmbuild -bb kernel-mppe-$maj_ver.spec --target "$arch--$os" --define "VERSION $maj_ver.$min_ver" --define "RELEASE $release" --define "BSD $BSD" --define "LICENSE $license" --quiet 1>$KMODLOG 2>$KMODLOG || exit 1
	rpm -U /usr/src/redhat/RPMS/$arch/kernel-mppe-$k_ver.$arch.rpm --justdb --replacefiles --replacepkgs
        echo "Done. RPM: /usr/src/redhat/RPMS/$arch/kernel-mppe-$k_ver.$arch.rpm"
    fi
fi

exit 0
